# @Time     : 2021/5/30
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com

def is_none(value):
    return value is None


def is_not_none(value):
    return value is not None



# @Time     : 2021/5/30
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from .reader import IniReader, JsonReader
from .builder import ReaderBuilder

from ..utils import Encoding, PathTypes


def create_ini_reader(file: PathTypes,
                      encoding: Encoding = None) -> IniReader:
    builder = ReaderBuilder(file, encoding)
    builder.add_ext_name(".int")
    return builder.build()


def create_json_reader(file: PathTypes,
                       encoding: Encoding = Encoding.UTF_8) -> JsonReader:
    builder = ReaderBuilder(file, encoding)
    builder.set_reader(JsonReader)
    builder.add_ext_name(".json")
    return builder.build()



# @Time     : 2021/5/27
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from .base import AbstractConfReader, IReaderBuilder
from .reader import IniReader
from ..utils import Encoding, ExtensionNameList, PathTypes
from ..validator.is_validators import is_none, is_string


class ReaderBuilder(IReaderBuilder):

    def __init__(self, file: PathTypes, encoding: Encoding = None):
        self._file = file
        self._ext_names: ExtensionNameList = []
        self._encoding = encoding
        self._reader_class = IniReader

    def set_reader(self, reader_class: AbstractConfReader) -> None:
        if not self._is_subclass(reader_class):
            raise ValueError(f"reader class need AbstractConfReader subclass, but got {type(reader_class).__name__}")
        self._reader_class = reader_class

    def add_ext_name(self, ext_name: str) -> None:
        if not is_string(ext_name):
            raise ValueError(f"file ext name need string, but got {type(ext_name).__name__}")
        self._ext_names.append(ext_name)

    def build(self):
        if is_none(self._reader_class):
            raise ValueError("reader class is not None")
        return self._reader_class(
            self._file, self._ext_names, self._encoding
        )

    def _is_subclass(self, reader):
        return issubclass(reader, AbstractConfReader)



# @Time     : 2021/5/27
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from enum import Enum
from typing import Any

from ..validator.is_validators import is_enum, is_enum_subclass


class EnumUtil(object):

    @classmethod
    def unenum_to_dict(cls, enum_subclass: Enum) -> dict:
        if not is_enum_subclass(enum_subclass):
            raise ValueError(f"enum klass not Enum subclass, got {type(enum_subclass).__name__}")
        unenum = cls.unenum
        return {member: unenum(member, "value") for _, member in enumerate(enum_subclass)}

    @classmethod
    def unenum(cls, any_or_member: Enum, name_or_value: str = "value"):
        method = {
            "name": cls._to_name,
            "value": cls._to_value
        }.get(name_or_value, "value")
        return method(any_or_member)

    @staticmethod
    def _to_value(any_or_member: Enum) -> Any:
        return any_or_member.value if is_enum(any_or_member) else any_or_member

    @staticmethod
    def _to_name(any_or_member: Enum) -> str:
        return any_or_member.name if is_enum(any_or_member) else any_or_member



# @Time     : 2021/5/28
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
try:
    import simplejson as json
except ImportError:
    import json
from collections import defaultdict
from configparser import ConfigParser
from typing import Dict, DefaultDict, List, Union
from uuid import uuid4

from .base import AbstractConfReader
from ..utils import (
    EncodingTypes,
    ExtensionNameList,
    PathTypes
)
from ..validator.is_validators import (
    is_dict,
    is_none,
    is_list,
    is_string
)


class IniReader(AbstractConfReader):
    DEFAULT = "DEFAULT"
    __file_ext__ = ".ini"

    def __init__(self,
                 file: PathTypes,
                 ext_names: ExtensionNameList = None,
                 encoding: EncodingTypes = None):
        super().__init__(file, ext_names, encoding)

        if ".ini" not in self._ext_names:
            self._ext_names.append(IniReader.__file_ext__)

        self._parser = self._init_parser(self.file)

    def get_float(self, node: str, option: str) -> float:
        node = self._get_node_from_section(node)
        result = self._parser.getfloat(node, self._get_option_from_section(node, option))
        return result

    def get_int(self, node: str, option: str) -> int:
        node = self._get_node_from_section(node)
        result = self._parser.getint(node, self._get_option_from_section(node, option))
        return result

    def get_string(self, node: str, option: str) -> str:
        node = self._get_node_from_section(node)
        result = self._parser.get(node, self._get_option_from_section(node, option))
        return self._strip_quota(result)

    def has_node(self, node: str):
        if is_none(node):
            return False
        return self._parser.has_section(node)

    def _get_option_from_section(self, node: str, option: str):
        result = option
        if not self._parser.has_option(node, result):
            raise ValueError("not found option {option}")
        return result

    def _get_node_from_section(self, node: str):
        return IniReader.DEFAULT if not self.has_node(node) else node

    def _strip_quota(self, result: str) -> str:
        if not is_string(result):
            raise ValueError(f"result need string, but got {type(result).__name__}")
        if not result:
            return result
        return result.strip("'").strip('"')

    def _init_parser(self, file: PathTypes) -> ConfigParser:
        parser = ConfigParser()
        self._read_ini(parser, file)
        return parser

    def _read_ini(self, parser: ConfigParser, file: PathTypes):
        if self.is_config_file(file):
            parser.read(file, encoding=self._to_encoding(self._encoding))


Number = Union[float, int]
JsonTypes = Union[bool, Number, Dict, List, None]


class JsonNode(object):

    def __init__(self, key: str, data: JsonTypes = None):
        self._key = key
        self._data = self._create_by_json(key, data)

    def get(self, option: str, default=None):
        return self._data.get(option, default)

    def has_option(self, option):
        return option in self._data

    def _create_by_json(self, key: str, data: JsonTypes) -> dict:
        if is_dict(data):
            return self._by_dict(key, data)
        elif is_list(data):
            return self._by_list(key, data)
        else:
            return self._by_value(key, data)

    def _by_dict(self, key, data: Dict):
        return {k: v for k, v in data.items()}

    def _by_list(self, key, data: List):
        return {index: value for index, value in enumerate(data)}

    def _by_value(self, key, data: Union[bool, None, Number]):
        return {key: data}


class JsonReader(AbstractConfReader):
    __file_ext__ = ".json"

    def __init__(self,
                 file: PathTypes,
                 ext_names: ExtensionNameList = None,
                 encoding: EncodingTypes = None):
        super().__init__(file, ext_names, encoding)

        if ".json" not in self._ext_names:
            self._ext_names.append(JsonReader.__file_ext__)

        self._parser = self._init_parser(self.file)

    def has_node(self, node: str) -> bool:
        return node in self._parser

    def get_float(self, node: str, option: str) -> float:
        return self._get_from_parser(node, option, self._to_float)

    def get_int(self, node: str, option: str) -> int:
        return self._get_from_parser(node, option, self._to_int)

    def get_string(self, node: str, option: str) -> str:
        return self._get_from_parser(node, option, self._to_string)

    def _get_from_parser(self, node, option: str, callback):
        json_node = self._get_node_from_parser(node)
        result = self._get_option_from_parser(json_node, option)
        return callback(result)

    def _get_option_from_parser(self, node: JsonNode, option: str):
        if not node.has_option(option):
            raise KeyError(f"not found option {option}")
        return node.get(option)

    def _get_node_from_parser(self, node):
        return self._parser[node]

    def _init_parser(self, file: PathTypes) -> DefaultDict[str, JsonNode]:
        data = self._read_json(file, self._encoding)
        filtered = self._filter_json(data)
        return defaultdict(self._default_node, filtered)

    def _filter_json(self, result: dict) -> Dict[str, JsonNode]:
        if not is_dict(result):
            return {}
        return {key: self._default_node(key, value) for key, value in result.items()}

    def _read_json(self, file, encoding: EncodingTypes = None) -> dict:
        result = {}
        if self.is_config_file(file):
            encoding = self._to_encoding(encoding) if not encoding else encoding
            with open(file, mode="r") as f:
                result = json.load(f, encoding=encoding)
        return result

    def _default_node(self, key: str = None, value=None):
        _key: str = key
        if _key is None:
            _key = uuid4()
        return JsonNode(key, value)



# @Time     : 2021/5/28
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
def is_int(value):
    return isinstance(value, int)


def is_float(value):
    return isinstance(value, float)


def is_number(value):
    return any([
        is_int(value),
        is_float(value)
    ])



# @Time     : 2021/5/28
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from .src import validator
from .src.validator import is_validators
from .src.utils import (
    Allowed,
    EnumUtil,
    Encoding,
    StringsUtil,
    StringOrBytesOrByteArray,
    PathUtil,
    PathTypes,
    ExtensionNameList
)
from .src.callback import AbstractCallbackManager, CallbackManager, AsyncCallbackManager
from .src.conf import (
    AbstractConfReader,
    IniReader,
    JsonReader,
    ReaderBuilder,
    create_ini_reader,
    create_json_reader
)



# @Time     : 2021/5/27
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from .base import AbstractConfReader, IReaderBuilder
from .reader import IniReader, JsonReader, JsonNode
from .builder import ReaderBuilder
from .api import create_ini_reader, create_json_reader



# @Time     : 2021/5/27
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
def is_string(value):
    return isinstance(value, str)


def is_bytes(value):
    return isinstance(value, bytes)


def is_bytearray(value):
    return isinstance(value, bytearray)


def is_any_string(value):
    return any([
        is_bytes(value),
        is_string(value)
    ])



# @Time     : 2021/5/26
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from enum import Enum
from typing import AnyStr, Union

from .enums import EnumUtil
from ..validator.is_validators import (
    is_bytes,
    is_bytearray,
    is_enum,
    is_any_string,
    is_string
)

StringOrBytesOrByteArray = Union[AnyStr, bytearray]
EncodingTypes = Union[AnyStr, Enum, None]


class Encoding(Enum):
    """
    默认编码方式
    """
    ASCII = "ascii"
    ISO_8859_1 = "ISO-8859-1"
    UTF_8 = "utf-8"


class StringsUtil(object):

    @classmethod
    def anystr_to_string(cls,
                         str_or_bytes: AnyStr,
                         encoding: EncodingTypes = None) -> str:
        """
        string or bytes to string
        :param str_or_bytes:
        :param encoding:
        :return:
        """
        string: str = ""
        if is_string(str_or_bytes):
            string = str_or_bytes
        elif is_bytes(str_or_bytes):
            string = str_or_bytes.decode(cls.to_encoding(encoding))
        else:
            raise ValueError(f"need a string or bytes, but got {type(str_or_bytes).__name__}")
        return string

    @classmethod
    def anystr_to_bytes(cls,
                        str_or_bytes_barray: StringOrBytesOrByteArray) -> bytes:
        """
        string, bytes, bytearray to bytes
        :param str_or_bytes_barray:
        :return:
        """
        bstring: bytes = b""
        if any([
            is_bytes(str_or_bytes_barray),
            is_bytearray(str_or_bytes_barray)
        ]):
            bstring = str_or_bytes_barray
        elif is_string(str_or_bytes_barray):
            bstring = str_or_bytes_barray.encode()
        else:
            raise ValueError(f"need string, bytes, bytearray, but got {type(str_or_bytes_barray).__name__}")
        return bstring

    @classmethod
    def to_encoding(cls, encoding: EncodingTypes) -> Union[str, None]:
        _encoding: str = ""
        if is_any_string(encoding):
            _encoding = StringsUtil.anystr_to_string(encoding)
        elif is_enum(encoding):
            _encoding = EnumUtil.unenum(encoding)
        else:
            _encoding = None
        return _encoding



# @Time     : 2021/5/27
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from abc import ABCMeta, abstractmethod


class AbstractValidator(metaclass=ABCMeta):

    def __init__(self, error_message: str):
        self._message = error_message

    @property
    def error_message(self):
        return self._message

    @abstractmethod
    def is_validate(self, value, **kwargs) -> bool:
        raise NotImplementedError("NotImplemented .is_validate(value) -> bool")

    @abstractmethod
    def raise_error(self, message: str):
        raise NotImplementedError("NotImplemented .raise_error(message)")

    def __call__(self, value, **kwargs):
        result = self.is_validate(value, **kwargs)
        if not result:
            self.raise_error(self.error_message)
        return result



# @Time     : 2021/5/27
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from typing import Callable, List, Union

from .base import AbstractValidator
from .is_validators import is_function

ValidatorTypes = Union[AbstractValidator, Callable]


class IValidatorManager(object):

    def add(self, validator: ValidatorTypes) -> int:
        raise NotImplementedError("NotImplemented .add() -> int")

    def remove(self, validator: ValidatorTypes) -> int:
        raise NotImplementedError("NotImplemented .remove() -> int")

    def is_validate(self, value, **kwargs) -> bool:
        raise NotImplementedError("NotImplemented .is_validate() -> bool")


class ValidatorManager(IValidatorManager):

    def __init__(self):
        self._validators: List[ValidatorTypes] = []

    @property
    def validators(self):
        return self._validators

    @property
    def length(self) -> int:
        return len(self.validators)

    def empty(self):
        return not self.length

    def add(self, validator):
        if not self._is_validator(validator):
            return self.length

        if not self._is_exists(validator):
            self.validators.append(validator)

        return self.length

    def remove(self, validator):
        if self.empty():
            return self.length
        idx = self._find(validator)
        if idx > -1:
            self.validators.pop(idx)
        return self.length

    def is_validate(self, value, **kwargs) -> bool:
        return all(validator(value) for validator in self)

    def __iter__(self):
        if not self.empty():
            for _, validator in enumerate(self.validators):
                yield validator

    def _is_validator(self, value):
        return any([
            is_function(value),
            isinstance(value, AbstractValidator)
        ])

    def _find(self, validator: ValidatorTypes) -> int:
        if not self._is_exists(validator):
            return -1
        return self.validators.index(validator)

    def _is_exists(self, validator: ValidatorTypes) -> bool:
        if self.empty():
            return False
        return validator in self.validators



# @Time     : 2021/5/30
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
"""
conf module interfaces
"""
from abc import ABCMeta, abstractmethod
from ..utils import EncodingTypes, ExtensionNameList, PathTypes, PathUtil, StringsUtil
from ..validator.is_validators import (
    is_any_string,
    is_float,
    is_int,
    is_list,
    is_none,
    is_number,
    is_string
)


class AbstractConfReader(metaclass=ABCMeta):
    """
    config reader abc class
    """

    def __init__(self,
                 file: PathTypes,
                 ext_names: ExtensionNameList = None,
                 encoding: EncodingTypes = None):
        self._file = file
        self._ext_names = [] if not is_list(ext_names) else ext_names
        self._encoding = encoding

    @property
    def file(self):
        return self._file

    @abstractmethod
    def has_node(self, node) -> bool:
        raise NotImplementedError("NotImplemented .has_node(node) -> bool")

    @abstractmethod
    def get_int(self, node: str, option) -> int:
        raise NotImplementedError("NotImplemented .get_int(node, option) -> int")

    @abstractmethod
    def get_float(self, node: str, option) -> float:
        raise NotImplementedError("NotImplemented .get_float(node, option) -> float")

    @abstractmethod
    def get_string(self, node: str, option) -> str:
        raise NotImplementedError("NotImplemented .get_string(node, option) -> str")

    def _to_int(self, value):
        if is_int(value):
            return value
        elif is_number(value):
            return int(value)
        else:
            raise ValueError(f"value need float or int, but got {type(value).__name__}")

    def _to_float(self, value):
        if is_float(value):
            return value
        elif is_number(value):
            return float(value)
        else:
            raise ValueError(f"value need float or int, but got {type(value).__name__}")

    def _to_string(self, value):
        if is_string(value):
            return value
        elif is_any_string(value):
            return StringsUtil.anystr_to_string(value)
        return str(value)

    def is_config_file(self, file: PathTypes) -> bool:
        return PathUtil.is_endswith_ext(file, self._ext_names)

    def _to_encoding(self, encoding: EncodingTypes = None):
        if is_none(encoding):
            return None
        return StringsUtil.to_encoding(encoding)


class IReaderBuilder(object):

    def set_reader(self, reader_class: AbstractConfReader) -> None:
        raise NotImplementedError("NotImplemented set_reader(reader) -> None")

    def add_ext_name(self, ext_name: str) -> None:
        raise NotImplementedError("NotImplemented .add_ext_name(ext_name) -> None")

    def build(self) -> AbstractConfReader:
        raise NotImplementedError("NotImplementedError .build() -> AbstractConfReader")
